with Ada.Calendar;
with Ada.Characters.Latin_1;
with Ada.Directories.Temporary;
with Ada.Environment_Variables;
with Ada.Exceptions;
with Ada.Streams.Stream_IO.Standard_Files;
with Ada.Strings.Unbounded;
with Ada.Text_IO.Text_Streams;
with Web.Lock_Files;
with Birdeyes.Configurations;
with Birdeyes.Debug;
with Birdeyes.LIRS;
with Birdeyes.Pages.Index_Page;
with Birdeyes.Pages.Refresh_Page;
with Birdeyes.Pages.Register_Page;
with Birdeyes.Pages.View_Page;
with Birdeyes.Syndications.Management;
with Birdeyes.URI;
with Birdeyes.Users.Management;
with Birdeyes.Users.View;
procedure Birdeyes.CGI is
	use type Ada.Calendar.Time;
	use type Ada.Strings.Unbounded.Unbounded_String;
	Line_Break : Character renames Ada.Characters.Latin_1.LF;
	Now : constant Ada.Calendar.Time := Ada.Calendar.Clock;
	Input : not null Ada.Streams.Stream_IO.Stream_Access :=
		Ada.Streams.Stream_IO.Stream (
			Ada.Streams.Stream_IO.Standard_Files.Standard_Input.all);
	Output : not null Ada.Streams.Stream_IO.Stream_Access :=
		Ada.Streams.Stream_IO.Stream (
			Ada.Streams.Stream_IO.Standard_Files.Standard_Output.all);
begin
	declare
		Temporary_Directory : String renames Configurations.Temporary_Directory;
	begin
		Ada.Directories.Create_Path (Temporary_Directory);
		Ada.Directories.Temporary.Set_Temporary_Directory (Temporary_Directory);
	end;
	if Configurations.Additional_PATH'Length /= 0 then
		Ada.Environment_Variables.Set (
			"PATH",
			Ada.Environment_Variables.Value ("PATH") & ":"
				& Configurations.Additional_PATH);
	end if;
	Debug.Hook (Configurations.Debug_Log_File_Name'Access, Now);
	Locked : declare
		Lock : Web.Lock_Files.Lock_Type :=
			Web.Lock_Files.Lock (Configurations.Lock_Name, Force => 60.0);
		-- application
		Syndication_Manager : Syndications.Management.Manager :=
			Syndications.Management.Create (Configurations.Syndications_Directory'Access);
		User_Manager : Users.Management.Manager :=
			Users.Management.Create (Configurations.Users_Directory'Access);
		-- HTTP Info
		Remote_Addr : constant String := Web.Remote_Addr;
		Remote_Host : constant String := Web.Remote_Host;
		Cookie : Web.Cookie := Web.Get_Cookie;
		Query_Strings : constant Web.Query_Strings := Web.Get_Query_Strings;
		Post : constant Boolean := Web.Post;
		Inputs : Web.Query_Strings renames Web.Get (Input);
		procedure Refresh_Page is
			Self_URI : constant String := Web.Request_URI;
		begin
			Web.Header_See_Other (Output, Self_URI);
			Web.Header_Content_Type (Output, Web.Text_HTML);
			Web.Header_Cookie (Output, Cookie, Now + Configurations.Cookie_Duration);
			Web.Header_Break (Output);
			Pages.Refresh_Page (Output, URI => Self_URI);
		end Refresh_Page;
		-- user
		User_Password : constant Users.Password_Digest :=
			Users.Digest (Pages.Get_User_Password (Cookie));
		User_Info : Users.User_Info;
		procedure Check_User (User_Id : in String) is
		begin
			if User_Id'Length /= 0 then
				Users.Management.Query (
					User_Manager,
					User_Id,
					User_Password,
					Remote_Addr,
					Remote_Host,
					Now,
					User_Info);
			end if;
		end Check_User;
		-- query
		Where : constant Pages.Page := Pages.Where (Query_Strings);
		View_User_Id : constant String :=
			Pages.Get_View_User_Id (Query_Strings, User_Manager);
		-- input
		Cmd : constant String := Pages.Get_Command (Inputs);
		Message : aliased Ada.Strings.Unbounded.Unbounded_String;
	begin
		if Web.Lock_Files.Forced (Lock) then
			Ada.Debug.Put ("forced to remove lock-file.");
		end if;
		if Post
			and then (Remote_Host'Length = 0 or else Remote_Host = Remote_Addr)
			and then Remote_Addr /= "127.0.0.1" -- localhost
			and then Remote_Addr /= "::1" -- IPv6 localhost
		then
			Web.Header_503 (Output);
			Web.Header_Break (Output);
		elsif Cmd = "logon" then
			Logon : declare
				New_User_Id : constant String := Pages.Get_New_User_Id (Inputs);
				New_User_Password : constant String := Pages.Get_New_User_Password (Inputs);
			begin
				if not Users.Valid_Id_String (New_User_Id) then
					Pages.Set_Message (Cookie, New_User_Id & " is not valid name for user id.");
					Refresh_Page;
				elsif not Users.Management.Exists (User_Manager, New_User_Id) then
					Web.Header_Content_Type (Output, Web.Text_HTML);
					Web.Header_Cookie (Output, Cookie, Now + Configurations.Cookie_Duration);
					Web.Header_Break (Output);
					Pages.Register_Page (
						Output,
						Template => Configurations.Template_Register,
						Base_Page => Where,
						View_User_Id => View_User_Id,
						New_User_Id => New_User_Id,
						New_User_Password => New_User_Password);
				else
					begin
						Users.Management.Query (
							User_Manager,
							New_User_Id,
							Users.Digest (New_User_Password),
							Remote_Addr,
							Remote_Host,
							Now,
							User_Info);
						Pages.Set_User (Cookie, New_User_Id, New_User_Password);
						Pages.Set_Message (Cookie, "You did logon as " & New_User_Id & ".");
						Refresh_Page;
					exception
						when Users.Management.Password_Mismatch =>
							Pages.Set_User (Cookie, "", "");
							Pages.Set_Message (Cookie, "Your inputed password is mismatch.");
							Refresh_Page;
					end;
				end if;
			end Logon;
		elsif Cmd = "logoff" then
			Logoff : declare
				User_Id : constant String := Pages.Get_User_Id (Cookie, User_Manager);
			begin
				if User_Id'Length /= 0 then
					Ada.Debug.Put ("logoff " & User_Id);
				else
					Ada.Debug.Put ("logoff (no user)");
				end if;
				Pages.Set_User (Cookie, "", "");
				Pages.Set_Message (Cookie, "You did logoff.");
				Refresh_Page;
				Debug.Clean_Temporary; -- addition
			end Logoff;
		elsif Cmd = "register" then
			Register : declare
				New_User_Id : constant String := Pages.Get_New_User_Id (Inputs);
				New_User_Password : constant String := Pages.Get_New_User_Password (Inputs);
				New_User_Password_2 : constant String :=
					Pages.Get_New_User_Confirmation_Password (Inputs);
			begin
				if New_User_Password /= New_User_Password_2 then
					Web.Header_Content_Type (Output, Web.Text_HTML);
					Web.Header_Cookie (Output, Cookie, Now + Configurations.Cookie_Duration);
					Web.Header_Break (Output);
					Pages.Set_Message (Cookie, "Your re-inputed password is mismatch.");
					Refresh_Page;
				else
					Users.Management.Add (
						User_Manager,
						New_User_Id,
						Users.Digest (New_User_Password),
						Remote_Addr => Users.Local_Host_Addr,
						Remote_Host => Users.Local_Host_Name,
						Time => Now,
						Info => User_Info);
					Pages.Set_User (Cookie, New_User_Id, New_User_Password);
					Pages.Set_Message (Cookie, "Success!");
					Refresh_Page;
				end if;
			end Register;
		else
			case Where is
				when Pages.Index =>
					if Cmd'Length /= 0 then
						Pages.Set_Message (Cookie, "An unknown command is send.");
						Refresh_Page;
					else
						declare
							User_Id : constant String := Pages.Get_User_Id (Cookie, User_Manager);
						begin
							Check_User (User_Id);
							Pages.Get_Message (Cookie, Message);
							Web.Header_Content_Type (Output, Web.Text_HTML);
							Web.Header_Cookie (Output, Cookie, Now + Configurations.Cookie_Duration);
							Web.Header_Break (Output);
							Pages.Index_Page (
								Output,
								Template => Configurations.Template_Index,
								Message => Message.Constant_Reference,
								User_Manager => User_Manager,
								User_Id => User_Id);
						end;
					end if;
				when Pages.View =>
					if Cmd = "add" then
						Add : declare
							User_Id : constant String := Pages.Get_User_Id (Cookie, User_Manager);
						begin
							declare
								Target_URI : constant String := URI.Normalize (Pages.Get_URI (Inputs));
								Showing : Users.Showing_Mode := Pages.Get_Showing (Inputs);
								Syndication_Info : Syndications.Syndication_Info;
								Articles : Syndications.Article_Lists.List;
							begin
								Users.Management.Query (
									User_Manager,
									User_Id,
									User_Password,
									Remote_Addr => Users.Local_Host_Addr,
									Remote_Host => Users.Local_Host_Name,
									Time => Now,
									Info => User_Info);
								if Users.Contains (User_Info.Watching, Target_URI) then
									Pages.Set_Message (Cookie, Target_URI & " is already watched.");
								else
									Syndications.Management.Get (
										Syndication_Manager,
										URI => Target_URI,
										Time => Now,
										Info => Syndication_Info,
										Articles => Articles);
									Users.Watching_Syndications.Append (
										User_Info.Watching,
										(URI => +Target_URI, Showing => Showing));
									Users.Management.Update (
										User_Manager,
										User_Id,
										User_Info);
									Pages.Set_Message (Cookie, Target_URI & " is added.");
								end if;
							exception
								when Syndications.Management.Not_Available =>
									Pages.Set_Message (Cookie, Target_URI & " is not available.");
							end;
							Refresh_Page;
						end Add;
					elsif Cmd = "remove" then
						Remove : declare
							User_Id : constant String := Pages.Get_User_Id (Cookie, User_Manager);
							Target_URI : constant String := Pages.Get_URI (Inputs);
						begin
							Users.Management.Query (
								User_Manager,
								User_Id,
								User_Password,
								Remote_Addr => Users.Local_Host_Addr,
								Remote_Host => Users.Local_Host_Name,
								Time => Now,
								Info => User_Info);
							if not Users.Contains (User_Info.Watching, Target_URI) then
								Pages.Set_Message (Cookie, Target_URI & " is not watched.");
							else
								Users.Delete (User_Info.Watching, Target_URI);
								Users.Management.Update (
									User_Manager,
									User_Id,
									User_Info);
								Pages.Set_Message (Cookie, Target_URI & " is removed.");
							end if;
							Refresh_Page;
						end Remove;
					elsif Cmd = "import" then
						Import : declare
							User_Id : constant String := Pages.Get_User_Id (Cookie, User_Manager);
							Message : aliased Ada.Strings.Unbounded.Unbounded_String;
							Success_List, Failure_List : aliased URI.Lists.List;
							Success_Count : Natural := 0;
						begin
							Users.Management.Query (
								User_Manager,
								User_Id,
								User_Password,
								Remote_Addr => Users.Local_Host_Addr,
								Remote_Host => Users.Local_Host_Name,
								Time => Now,
								Info => User_Info);
							declare
								Temporary_LIRS : constant String :=
									Ada.Directories.Temporary.Create_Temporary_File;
								File : Ada.Text_IO.File_Type;
							begin
								Ada.Text_IO.Create (File, Name => Temporary_LIRS);
								Ada.Text_IO.Put (File, Pages.Get_Content (Inputs));
								Ada.Text_IO.Close (File);
								Ada.Text_IO.Open (File, Ada.Text_IO.In_File, Name => Temporary_LIRS);
								LIRS.Import (
									Syndication_Manager,
									File,
									Now,
									Success_List,
									Failure_List);
								Ada.Text_IO.Delete (File);
							end;
							Ada.Strings.Unbounded.Append (Message, "Success!");
							for I in Success_List.Iterate loop
								declare
									URI : Ada.Strings.Unbounded.Unbounded_String
										renames Success_List.Constant_Reference (I);
								begin
									if Users.Contains (User_Info.Watching, URI.Constant_Reference) then
										Ada.Strings.Unbounded.Append (
											Message,
											Line_Break & URI.Constant_Reference & " is already watched.");
									else
										Users.Watching_Syndications.Append (
											User_Info.Watching,
											(URI => URI, Showing => Users.Default_Watching_Syndication.Showing));
										Success_Count := Success_Count + 1;
									end if;
								end;
							end loop;
							for I in Failure_List.Iterate loop
								declare
									URI : Ada.Strings.Unbounded.Unbounded_String
										renames Failure_List.Constant_Reference (I);
								begin
									Ada.Strings.Unbounded.Append (
										Message,
										Line_Break & URI.Constant_Reference & " is not available.");
								end;
							end loop;
							Users.Management.Update (
								User_Manager,
								User_Id,
								User_Info);
							Ada.Strings.Unbounded.Append (
								Message,
								Line_Break & Natural'Image (Success_Count) & " feeds are imported.");
							Pages.Set_Message (Cookie, Message.Constant_Reference);
							Refresh_Page;
						end Import;
					elsif Cmd = "export" then
						Export : declare
							Watching : aliased Users.Watching_Syndications.List;
							File : Ada.Text_IO.File_Type;
						begin
							Users.Management.Query (
								User_Manager,
								View_User_Id,
								Watching => Watching);
							Web.Header_Content_Type (Output, Web.Text_Plain);
							Web.Header_Break (Output);
							Ada.Text_IO.Text_Streams.Open (File, Ada.Text_IO.Out_File, Output);
							LIRS.Export (
								Syndication_Manager,
								File,
								Users.To_URI_List (Watching),
								Agent_URI => Web.Compose (Web.HTTP, Web.Host, Web.Request_Path));
						end Export;
					elsif Cmd'Length /= 0 then
						Pages.Set_Message (Cookie, "An unknown command is send.");
						Refresh_Page;
					else
						declare
							User_Id : constant String := Pages.Get_User_Id (Cookie, User_Manager);
						begin
							Check_User (User_Id);
							declare
								Watching : aliased Users.Watching_Syndications.List;
								URI_List : URI.Lists.List;
								Update_URI_List : aliased URI.Lists.List;
								Watched : aliased Syndications.Syndication_Info_And_Articles_Lists.List;
							begin
								Users.Management.Query (
									User_Manager,
									View_User_Id,
									Watching => Watching);
								URI_List := Users.To_URI_List (Watching);
								if not Pages.Has_Message (Cookie) then
									Update_URI_List := Syndications.Management.Select_For_Update (
										Syndication_Manager,
										URI_List);
								end if;
								Syndications.Management.Get (
									Syndication_Manager,
									URI_List,
									Update_URI_List,
									Now,
									Watched);
								Users.View.Sort (Watching, Watched, Now);
								Pages.Get_Message (Cookie, Message);
								Web.Header_Content_Type (Output, Web.Text_HTML);
								Web.Header_Cookie (Output, Cookie, Now + Configurations.Cookie_Duration);
								Web.Header_Break (Output);
								Pages.View_Page (
									Output,
									Template => Configurations.Template_View,
									Message => Message.Constant_Reference,
									View_User_Id => View_User_Id,
									Watching => Watching,
									Watched_Syndications => Watched,
									User_Id => User_Id);
							end;
						end;
					end if;
				when Pages.RSS =>
					declare
						Watching : aliased Users.Watching_Syndications.List;
						URI_List : URI.Lists.List;
						Update_URI_List : aliased URI.Lists.List;
						Watched : aliased Syndications.Syndication_Info_And_Articles_Lists.List;
						Articles : aliased Syndications.Article_Lists.List;
						View_URI : aliased Ada.Strings.Unbounded.Unbounded_String;
						procedure Add_To_View_URI (S : in String) is
						begin
							Ada.Strings.Unbounded.Append (View_URI, S);
						end Add_To_View_URI;
						procedure Add_Query is new Web.Generic_Write_Query_In_Attribute (
							Add_To_View_URI,
							Web.XHTML);
					begin
						Add_To_View_URI (Web.Compose (Web.HTTP, Web.Host, Web.Request_Path));
						Add_Query (Pages.Parameters_To_View_Page (View_User_Id));
						Users.Management.Query (
							User_Manager,
							View_User_Id,
							Watching => Watching);
						URI_List := Users.To_URI_List (Watching);
						Update_URI_List := Syndications.Management.Select_For_Update (
							Syndication_Manager,
							URI_List);
						Syndications.Management.Get (
							Syndication_Manager,
							URI_List,
							Update_URI_List,
							Now,
							Watched);
						Users.View.Fold (Watching, Watched, Now, Articles);
						Web.Header_Content_Type (Output, Web.Application_RSS_XML);
						Web.Header_Break (Output);
						Users.View.Write_RSS (
							Output,
							Title => View_User_Id & "'s antenna",
							Link => View_URI.Constant_Reference,
							Articles => Articles);
					end;
			end case;
		end if;
	end Locked;
exception
	when Web.Lock_Files.Lock_Error =>
		Web.Header_Content_Type (Output, Web.Text_Plain);
		Web.Header_Break (Output);
		declare
			Message : constant String := "White fog. Wait 1 minute!";
		begin
			Ada.Debug.Put (Message);
			String'Write (Output, Message);
			Character'Write (Output, Line_Break);
		end;
	when E : others =>
		Web.Header_Content_Type (Output, Web.Text_Plain);
		Web.Header_Break (Output);
		declare
			Message : constant String := Ada.Exceptions.Exception_Information (E);
		begin
			Ada.Debug.Put (Message);
			String'Write (Output, Message);
			Character'Write (Output, Line_Break);
		end;
end Birdeyes.CGI;
