with Ada.Calendar.Formatting;
with Ada.Characters.Latin_1;
with Ada.Command_Line;
with Ada.Directories.Temporary;
with Ada.Exceptions;
with Ada.Hierarchical_File_Names;
with Ada.Strings.Functions;
with Ada.Strings.Unbounded;
with Ada.Text_IO.Text_Streams;
with Birdeyes.Configurations;
with Birdeyes.LIRS;
with Birdeyes.Syndications.Management;
with Birdeyes.URI;
with Birdeyes.Users.Management;
with Birdeyes.Users.View;
procedure Birdeyes.Command is
	use type Ada.Calendar.Time;
	use type Ada.Strings.Unbounded.Unbounded_String;
	package Latin_1 renames Ada.Characters.Latin_1;
	Now : constant Ada.Calendar.Time := Ada.Calendar.Clock;
	Base_Directory : constant String :=
		Ada.Hierarchical_File_Names.Unchecked_Containing_Directory (
			Ada.Command_Line.Command_Name);
begin
	declare
		Temporary_Directory : constant String :=
			Ada.Hierarchical_File_Names.Compose (
				Base_Directory,
				Configurations.Temporary_Directory);
	begin
		Ada.Directories.Create_Path (Temporary_Directory);
		Ada.Directories.Temporary.Set_Temporary_Directory (Temporary_Directory);
	end;
	declare
		Command_Line_Error : exception;
		-- application
		Syndications_Directory : aliased constant String :=
			Ada.Hierarchical_File_Names.Compose (
				Base_Directory,
				Configurations.Syndications_Directory);
		Users_Directory : aliased constant String :=
			Ada.Hierarchical_File_Names.Compose (
				Base_Directory,
				Configurations.Users_Directory);
		Syndication_Manager : Syndications.Management.Manager :=
			Syndications.Management.Create (Syndications_Directory'Unchecked_Access);
		User_Manager : Users.Management.Manager :=
			Users.Management.Create (Users_Directory'Unchecked_Access);
	begin
		if Ada.Command_Line.Argument_Count = 0 then
			Ada.Text_IO.Put_Line (
				Ada.Text_IO.Standard_Error,
				"Please, specify subcommand.");
		else
			declare
				function Find_Separator (S : String) return Positive is
					Result : Natural := Ada.Strings.Functions.Index_Element_Forward (S, ':');
				begin
					if Result = 0 then
						return S'Last + 1;
					else
						return Result;
					end if;
				end Find_Separator;
				Subcommand : constant String := Ada.Command_Line.Argument (1);
			begin
				if Subcommand = "--help" then
					Ada.Text_IO.Put_Line ("birdeyes <subcommand> [args]");
					Ada.Text_IO.New_Line;
					Ada.Text_IO.Put_Line ("Available subcommands:");
					Ada.Text_IO.Put_Line (Latin_1.HT & "register USER:PASSWORD");
					Ada.Text_IO.Put_Line (Latin_1.HT & "add USER:PASSWORD URI [latest|expanded]");
					Ada.Text_IO.Put_Line (Latin_1.HT & "remove USER:PASSWORD URI");
					Ada.Text_IO.Put_Line (Latin_1.HT & "import USER:PASSWORD < LIRS");
					Ada.Text_IO.Put_Line (Latin_1.HT & "export USER > LIRS");
					Ada.Text_IO.Put_Line (Latin_1.HT & "update USER [URI]");
					Ada.Text_IO.Put_Line (Latin_1.HT & "summary USER");
					Ada.Text_IO.Put_Line (Latin_1.HT & "rss USER");
				elsif Subcommand = "register" then
					if Ada.Command_Line.Argument_Count < 2 then
						raise Command_Line_Error with "User id and password are required.";
					end if;
					declare
						User_Id_And_Password : constant String := Ada.Command_Line.Argument (2);
						Separator : constant Positive := Find_Separator (User_Id_And_Password);
						User_Id : constant String :=
							User_Id_And_Password (User_Id_And_Password'First .. Separator - 1);
						User_Password : Users.Password_Digest
							renames Users.Digest (
								User_Id_And_Password (Separator + 1 .. User_Id_And_Password'Last));
					begin
						if not Users.Valid_Id_String (User_Id) then
							raise Command_Line_Error with User_Id & " is not valid name for user id.";
						elsif Separator >= User_Id_And_Password'Last then
							Ada.Text_IO.Put_Line (
								Ada.Text_IO.Standard_Error,
								"warning: Specified password is empty.");
						end if;
						if Users.Management.Exists (User_Manager, User_Id) then
							raise Command_Line_Error with User_Id & " is already existing.";
						end if;
						declare
							Info : Users.User_Info;
						begin
							Users.Management.Add (
								User_Manager,
								User_Id,
								User_Password,
								Remote_Addr => Users.Local_Host_Addr,
								Remote_Host => Users.Local_Host_Name,
								Time => Now,
								Info => Info);
						end;
					end;
				elsif Subcommand = "add" then
					if Ada.Command_Line.Argument_Count < 3 then
						raise Command_Line_Error with "User id, password and URI are required.";
					end if;
					declare
						User_Id_And_Password : constant String := Ada.Command_Line.Argument (2);
						Separator : constant Positive := Find_Separator (User_Id_And_Password);
						User_Id : constant String :=
							Users.Management.Registered_Id (
								User_Manager,
								User_Id_And_Password (User_Id_And_Password'First .. Separator - 1));
						User_Password : Users.Password_Digest
							renames Users.Digest (
								User_Id_And_Password (Separator + 1 .. User_Id_And_Password'Last));
						Target_URI : constant String := URI.Normalize (Ada.Command_Line.Argument (3));
						Showing : Users.Showing_Mode := Users.Latest;
					begin
						if Ada.Command_Line.Argument_Count >= 4 then
							declare
								S : constant String := Ada.Command_Line.Argument (4);
							begin
								Showing := Users.Showing_Mode'Value (S);
							exception
								when Constraint_Error =>
									raise Command_Line_Error with S & " is not showing mode. (latest or expanded)";
							end;
						end if;
						declare
							User_Info : Users.User_Info;
							Syndication_Info : Syndications.Syndication_Info;
							Articles : Syndications.Article_Lists.List;
						begin
							Users.Management.Query (
								User_Manager,
								User_Id,
								User_Password,
								Remote_Addr => Users.Local_Host_Addr,
								Remote_Host => Users.Local_Host_Name,
								Time => Now,
								Info => User_Info);
							if Users.Contains (User_Info.Watching, Target_URI) then
								Ada.Text_IO.Put_Line (
									Ada.Text_IO.Standard_Error,
									Target_URI & " is already watched.");
							else
								Syndications.Management.Get (
									Syndication_Manager,
									URI => Target_URI,
									Time => Now,
									Info => Syndication_Info,
									Articles => Articles);
								Users.Watching_Syndications.Append (
									User_Info.Watching,
									(URI => +Target_URI, Showing => Showing));
								Users.Management.Update (
									User_Manager,
									User_Id,
									User_Info);
							end if;
						exception
							when Syndications.Management.Not_Available =>
								Ada.Text_IO.Put_Line (
									Ada.Text_IO.Standard_Error,
									Target_URI & " is not available.");
						end;
					end;
				elsif Subcommand = "remove" then
					if Ada.Command_Line.Argument_Count < 3 then
						raise Command_Line_Error with "User id, password and URI are required.";
					end if;
					declare
						User_Id_And_Password : constant String := Ada.Command_Line.Argument (2);
						Separator : constant Positive := Find_Separator (User_Id_And_Password);
						User_Id : constant String :=
							Users.Management.Registered_Id (
								User_Manager,
								User_Id_And_Password (User_Id_And_Password'First .. Separator - 1));
						User_Password : Users.Password_Digest
							renames Users.Digest (
								User_Id_And_Password (Separator + 1 .. User_Id_And_Password'Last));
						URI : constant String := Ada.Command_Line.Argument (3);
					begin
						declare
							User_Info : Users.User_Info;
						begin
							Users.Management.Query (
								User_Manager,
								User_Id,
								User_Password,
								Remote_Addr => Users.Local_Host_Addr,
								Remote_Host => Users.Local_Host_Name,
								Time => Now,
								Info => User_Info);
							if not Users.Contains (User_Info.Watching, URI) then
								Ada.Text_IO.Put_Line (Ada.Text_IO.Standard_Error, URI & " is not watched.");
							else
								Users.Delete (User_Info.Watching, URI);
								Users.Management.Update (
									User_Manager,
									User_Id,
									User_Info);
							end if;
						end;
					end;
				elsif Subcommand = "import" then
					if Ada.Command_Line.Argument_Count < 2 then
						raise Command_Line_Error with "User id, password are required.";
					end if;
					declare
						User_Id_And_Password : constant String := Ada.Command_Line.Argument (2);
						Separator : constant Positive := Find_Separator (User_Id_And_Password);
						User_Id : constant String :=
							Users.Management.Registered_Id (
								User_Manager,
								User_Id_And_Password (User_Id_And_Password'First .. Separator - 1));
						User_Password : Users.Password_Digest
							renames Users.Digest (
								User_Id_And_Password (Separator + 1 .. User_Id_And_Password'Last));
					begin
						declare
							User_Info : Users.User_Info;
							Success_List, Failure_List : aliased URI.Lists.List;
						begin
							Users.Management.Query (
								User_Manager,
								User_Id,
								User_Password,
								Remote_Addr => Users.Local_Host_Addr,
								Remote_Host => Users.Local_Host_Name,
								Time => Now,
								Info => User_Info);
							LIRS.Import (
								Syndication_Manager,
								Ada.Text_IO.Standard_Input.all,
								Now,
								Success_List,
								Failure_List);
							for I in Success_List.Iterate loop
								declare
									URI : Ada.Strings.Unbounded.Unbounded_String
										renames Success_List.Constant_Reference (I);
								begin
									if Users.Contains (User_Info.Watching, URI.Constant_Reference) then
										Ada.Text_IO.Put_Line (
											Ada.Text_IO.Standard_Error,
											URI.Constant_Reference & " is already watched.");
									else
										Users.Watching_Syndications.Append (
											User_Info.Watching,
											(URI => URI, Showing => Users.Default_Watching_Syndication.Showing));
									end if;
								end;
							end loop;
							for I in Failure_List.Iterate loop
								declare
									URI : Ada.Strings.Unbounded.Unbounded_String
										renames Failure_List.Constant_Reference (I);
								begin
									Ada.Text_IO.Put_Line (
										Ada.Text_IO.Standard_Error,
										URI.Constant_Reference & " is not available.");
								end;
							end loop;
							Users.Management.Update (
								User_Manager,
								User_Id,
								User_Info);
						end;
					end;
				elsif Subcommand = "export" then
					if Ada.Command_Line.Argument_Count < 2 then
						raise Command_Line_Error with "User id is required.";
					end if;
					declare
						User_Id : constant String :=
							Users.Management.Registered_Id (User_Manager, Ada.Command_Line.Argument (2));
					begin
						declare
							Watching : aliased Users.Watching_Syndications.List;
						begin
							Users.Management.Query (
								User_Manager,
								User_Id,
								Watching => Watching);
							LIRS.Export (
								Syndication_Manager,
								Ada.Text_IO.Standard_Output.all,
								Users.To_URI_List (Watching),
								Agent_URI => "");
						end;
					end;
				elsif Subcommand = "update" then
					if Ada.Command_Line.Argument_Count < 2 then
						raise Command_Line_Error with "User id is required.";
					end if;
					declare
						User_Id : constant String :=
							Users.Management.Registered_Id (User_Manager, Ada.Command_Line.Argument (2));
					begin
						declare
							Watching : aliased Users.Watching_Syndications.List;
							URI_List : URI.Lists.List;
							Update_URI_List : aliased URI.Lists.List;
							Watched : aliased Syndications.Syndication_Info_And_Articles_Lists.List;
						begin
							Users.Management.Query (
								User_Manager,
								User_Id,
								Watching => Watching);
							URI_List := Users.To_URI_List (Watching);
							if Ada.Command_Line.Argument_Count >= 3 then
								declare
									Update_URI : constant String := Ada.Command_Line.Argument (3);
								begin
									if not Users.Contains (Watching, Update_URI) then
										raise Command_Line_Error with Update_URI & " is not watched.";
									else
										URI.Lists.Append (Update_URI_List, +Update_URI);
									end if;
								end;
							else
								Update_URI_List := Syndications.Management.Select_For_Update (
									Syndication_Manager,
									URI_List);
							end if;
							Syndications.Management.Get (
								Syndication_Manager,
								URI_List,
								Update_URI_List,
								Now,
								Watched);
							for I in Watched.Iterate loop
								declare
									Watched_Syndication : Syndications.Syndication_Info_And_Articles
										renames Watched.Constant_Reference (I);
								begin
									if Watched_Syndication.Info.Last_Tried_Time = Now then
										if Watched_Syndication.Info.Last_Avail_Time < Now then
											Ada.Text_IO.Put_Line (
												Ada.Text_IO.Standard_Error,
												Watched_Syndication.Info.URI.Constant_Reference & " is not available.");
										elsif Watched_Syndication.Info.Last_Detected_Time = Now then
											Ada.Text_IO.Put_Line (
												Watched_Syndication.Info.URI.Constant_Reference & " is updated.");
										else
											Ada.Text_IO.Put_Line (
												Watched_Syndication.Info.URI.Constant_Reference & " is not updated.");
										end if;
									end if;
								end;
							end loop;
						end;
					end;
				elsif Subcommand = "summary" then
					if Ada.Command_Line.Argument_Count < 2 then
						raise Command_Line_Error with "User id is required.";
					end if;
					declare
						User_Id : constant String :=
							Users.Management.Registered_Id (User_Manager, Ada.Command_Line.Argument (2));
					begin
						declare
							Watching : aliased Users.Watching_Syndications.List;
							Watched : aliased Syndications.Syndication_Info_And_Articles_Lists.List;
						begin
							Users.Management.Query (
								User_Manager,
								User_Id,
								Watching => Watching);
							Syndications.Management.Get (
								Syndication_Manager,
								Users.To_URI_List (Watching),
								URI.Lists.Empty_List,
								Now,
								Watched);
							for I in Watched.Iterate loop
								declare
									Watched_Syndication : Syndications.Syndication_Info_And_Articles
										renames Watched.Constant_Reference (I);
								begin
									Ada.Text_IO.Put_Line (
										"- uri: " & Watched_Syndication.Info.URI.Constant_Reference);
									Ada.Text_IO.Put_Line (
										"  version: "
											& Syndications.Syndication_Version'Image (Watched_Syndication.Info.Version));
									Ada.Text_IO.Put_Line (
										"  link: " & Watched_Syndication.Info.Link.Constant_Reference);
									Ada.Text_IO.Put_Line (
										"  title: " & Watched_Syndication.Info.Title.Constant_Reference);
									Ada.Text_IO.Put_Line (
										"  author: " & Watched_Syndication.Info.Author.Constant_Reference);
									Ada.Text_IO.Put_Line (
										"  last-modified-time: "
											& Ada.Calendar.Formatting.Image (
												Syndications.Last_Modified_Time (Watched_Syndication.Info)));
									Ada.Text_IO.Put_Line ("  articles:");
									for J in Watched_Syndication.Articles.Iterate loop
										declare
											Article : Syndications.Article
												renames Watched_Syndication.Articles.Constant_Reference (J);
										begin
											Ada.Text_IO.Put_Line ("  - link: " & Article.Link.Constant_Reference);
											Ada.Text_IO.Put_Line ("    title: " & Article.Title.Constant_Reference);
											Ada.Text_IO.Put_Line ("    author: " & Article.Author.Constant_Reference);
											Ada.Text_IO.Put_Line (
												"    time: " & Ada.Calendar.Formatting.Image (Article.Time));
										end;
									end loop;
								end;
							end loop;
						end;
					end;
				elsif Subcommand = "rss" then
					if Ada.Command_Line.Argument_Count < 2 then
						raise Command_Line_Error with "User id is required.";
					end if;
					declare
						User_Id : constant String :=
							Users.Management.Registered_Id (User_Manager, Ada.Command_Line.Argument (2));
					begin
						declare
							Watching : aliased Users.Watching_Syndications.List;
							Watched : aliased Syndications.Syndication_Info_And_Articles_Lists.List;
							Articles : aliased Syndications.Article_Lists.List;
						begin
							Users.Management.Query (
								User_Manager,
								User_Id,
								Watching => Watching);
							Syndications.Management.Get (
								Syndication_Manager,
								Users.To_URI_List (Watching),
								URI.Lists.Empty_List,
								Now,
								Watched);
							Users.View.Fold (Watching, Watched, Now, Articles);
							Users.View.Write_RSS (
								Ada.Text_IO.Text_Streams.Stream (Ada.Text_IO.Standard_Output.all),
								Articles => Articles);
						end;
					end;
				else
					raise Command_Line_Error with Subcommand & " is unknown subcommand.";
				end if;
			end;
		end if;
	exception
		when E : Command_Line_Error =>
			Ada.Text_IO.Put_Line (
				Ada.Text_IO.Standard_Error,
				Ada.Exceptions.Exception_Message (E));
	end;
end Birdeyes.Command;
