with Ada.Calendar.Formatting;
with Ada.Calendar.Time_Zones;
with Ada.Directories.Temporary;
with Ada.Hierarchical_File_Names;
with Ada.Streams.Stream_IO;
with System.Debug;
with Ada.Characters.Latin_1;
package body Birdeyes.Debug is
	
	Line_Break : Character renames Ada.Characters.Latin_1.LF;
	
	Name : Static_String_Access;
	File : Ada.Streams.Stream_IO.File_Type;
	Time : Ada.Calendar.Time;
	
	procedure Start is
	begin
		if Name = null then
			raise Program_Error with "debug log handler is not installed.";
		end if;
		-- create the directory
		declare
			Dir : constant String :=
				Ada.Hierarchical_File_Names.Unchecked_Containing_Directory (Name.all);
		begin
			if Dir'Length /= 0 then
				Ada.Directories.Create_Path (Dir);
			end if;
		end;
		-- open
		Ada.Streams.Stream_IO.Create (
			File,
			Ada.Streams.Stream_IO.Append_File,
			Name => Name.all);
		declare
			Time_Offset : constant Ada.Calendar.Time_Zones.Time_Offset :=
				Ada.Calendar.Time_Zones.UTC_Time_Offset;
			Time_Image : constant String :=
				Ada.Calendar.Formatting.Image (Time, Time_Zone => Time_Offset);
			Offset_Image : constant String :=
				Ada.Calendar.Formatting.Image (Time_Offset);
			Stream : Ada.Streams.Stream_IO.Stream_Access :=
				Ada.Streams.Stream_IO.Stream (File);
		begin
			String'Write (
				Stream,
				"---- " & Time_Image & " (GMT" & Offset_Image & ") ----" & Line_Break);
		end;
	end Start;
	
	function Put (
		S : in String;
		Source_Location : in String;
		Enclosing_Entity : in String)
		return Boolean is
	begin
		if not Ada.Streams.Stream_IO.Is_Open (File) then
			Start;
		end if;
		declare
			Stream : Ada.Streams.Stream_IO.Stream_Access :=
				Ada.Streams.Stream_IO.Stream (File);
		begin
			String'Write (
				Stream,
				Source_Location & ": (" & Enclosing_Entity & ") " & S & Line_Break);
		end;
		Ada.Streams.Stream_IO.Flush (File);
		return True;
	end Put;
	
	-- implementation
	
	procedure Hook (
		Name : not null Static_String_Access;
		Time : in Ada.Calendar.Time) is
	begin
		Debug.Name := Name;
		Debug.Time := Time;
		System.Debug.Put_Hook := Put'Access;
	end Hook;
	
	procedure Clean_Temporary is
		Search : aliased Ada.Directories.Search_Type;
	begin
		Ada.Directories.Start_Search (Search,
			Directory => Ada.Directories.Temporary.Temporary_Directory,
			Filter => (others => True));
		while Ada.Directories.More_Entries(Search) loop
			declare
				Directory_Entry : Ada.Directories.Directory_Entry_Type
					renames Ada.Directories.Look_Next_Entry (Search);
				Full_Name : constant String :=
					Ada.Directories.Full_Name (Directory_Entry);
			begin
				case Ada.Directories.Kind (Directory_Entry) is
					when Ada.Directories.Directory =>
						Ada.Directories.Delete_Tree (Full_Name);
					when Ada.Directories.Ordinary_File
						| Ada.Directories.Special_File =>
						Ada.Directories.Delete_File (Full_Name);
				end case;
			end;
			Ada.Directories.Skip_Next_Entry (Search);
		end loop;
		Ada.Directories.End_Search (Search);
	end Clean_Temporary;
	
end Birdeyes.Debug;
