with Ada.Characters.ASCII.Handling;
with Ada.Command_Line;
with Ada.Directories.Temporary;
with Ada.Environment_Variables;
with Ada.Exceptions;
with Ada.Hierarchical_File_Names;
with Ada.Streams.Stream_IO;
with Ada.Strings.Functions;
with Birdeyes.Calendar;
package body Birdeyes.HTTP is
	use type Ada.Calendar.Time;
	use type Ada.Command_Line.Exit_Status;
	use type Ada.Streams.Stream_Element_Array;
	use type Ada.Streams.Stream_Element_Offset;
	use type Ada.Strings.Unbounded.Unbounded_String;
	
	Gzip_Signature : constant Ada.Streams.Stream_Element_Array :=
		(16#1f#, 16#8b#, 16#08#, 16#00#);
	
	function Is_Gzip (Name : String) return Boolean is
		File : Ada.Streams.Stream_IO.File_Type :=
			Ada.Streams.Stream_IO.Open (Ada.Streams.Stream_IO.In_File, Name => Name);
		Signature : Ada.Streams.Stream_Element_Array (1 .. 4);
		Last : Ada.Streams.Stream_Element_Count;
	begin
		Ada.Streams.Read (Ada.Streams.Stream_IO.Stream (File).all, Signature, Last);
		Ada.Streams.Stream_IO.Close (File);
		return Last = Gzip_Signature'Length and then Signature = Gzip_Signature;
	end Is_Gzip;
	
	procedure Rename_Or_Extract (Source, Target : in String) is
	begin
		if Is_Gzip (Source) then
			declare
				Source_2 : constant String := Source & ".gz";
				Modification_Time : constant Ada.Calendar.Time :=
					Ada.Directories.Modification_Time (Source);
				Gzip : Ada.Processes.Process;
				Command : Ada.Processes.Command_Type;
				Status : Ada.Command_Line.Exit_Status;
			begin
				Ada.Directories.Rename (Source, Source_2, Overwrite => False);
				Ada.Processes.Append (Command, "/usr/bin/env");
				Ada.Processes.Append (Command, "gzip");
				Ada.Processes.Append (Command, "-d");
				Ada.Processes.Append (Command, Source_2);
				Ada.Debug.Put (Ada.Processes.Image (Command));
				Ada.Processes.Create (Gzip, Command);
				Ada.Processes.Wait (Gzip, Status);
				if Status = 127 then
					raise Program_Error with "Gzip is not installed.";
				elsif Status /= 0 then
					raise Data_Error;
				end if;
				Ada.Directories.Set_Modification_Time (
					Source,
					Modification_Time);
				Ada.Directories.Rename (Source, Target, Overwrite => True);
			end;
		else
			Ada.Directories.Rename (Source, Target, Overwrite => True);
		end if;
	end Rename_Or_Extract;
	
	-- emulate file name encoding of Wget
	function Downloaded_Name (URI : String) return String is
		Result : Ada.Strings.Unbounded.Unbounded_String;
		I, Q : Natural;
	begin
		I := Ada.Strings.Functions.Index_Element_Forward (URI, ':');
		if I = 0 then
			I := URI'First;
		else
			I := I + 1;
		end if;
		for J in 1 .. 2 loop
			if I <= URI'Last and then URI (I) = '/' then
				I := I + 1;
			end if;
		end loop;
		Q := Ada.Strings.Functions.Index_Element_Forward (URI (I .. URI'Last), '?');
		if Q = 0 then
			Q := URI'Last;
		end if;
		I := Ada.Strings.Functions.Index_Element_Backward (URI (I .. Q), '/');
		if I = 0
			or else I = URI'Last
			or else URI (I + 1 .. URI'Last) = "."
			or else URI (I + 1 .. URI'Last) = ".."
		then
			Ada.Strings.Unbounded.Append (Result, "index.html");
				-- default name of wget --default-page
			I := 0;
		elsif I < URI'Last and then URI (I + 1) = '?' then
			Ada.Strings.Unbounded.Append (Result, "index.html");
			I := I + 1; -- skip '/'
		else
			I := I + 1; -- skip '/'
		end if;
		if I > 0 then
			while I <= URI'Last loop
				declare
					C : Character;
				begin
					case URI (I) is
						when '%' =>
							if I <= URI'Last - 2
								and then Ada.Characters.ASCII.Handling.Is_Alphanumeric (URI (I + 1))
								and then Ada.Characters.ASCII.Handling.Is_Alphanumeric (URI (I + 2))
							then
								C := Character'Value ("Hex_" & URI (I + 1 .. I + 2));
								I := I + 3;
							else
								C := '%';
								I := I + 1;
							end if;
						when others =>
							C := URI (I);
							I := I + 1;
					end case;
					case C is
						when '/' =>
							Ada.Strings.Unbounded.Append (Result, "%2F");
						when others =>
							Ada.Strings.Unbounded.Append (Result, C);
					end case;
				end;
			end loop;
		end if;
		return Ada.Strings.Unbounded.To_String (Result);
	end Downloaded_Name;
	
	-- implementation
	
	procedure Start (
		Object : in out Connection;
		URI : in String;
		Name : in String)
	is
		Command : Ada.Processes.Command_Type;
	begin
		Object.Temporary_Directory :=
			+Ada.Directories.Temporary.Create_Temporary_Directory;
		Object.Temporary_Name := +Ada.Hierarchical_File_Names.Compose (
			Object.Temporary_Directory.Constant_Reference,
			Downloaded_Name (URI));
		Object.Name := +Name;
		Ada.Processes.Append (Command, "/usr/bin/env");
		Ada.Processes.Append (Command, "wget");
		if Ada.Directories.Exists (Name) then
			Object.Saved_Time := Ada.Directories.Modification_Time (Name);
			Object.Saved_Size := Ada.Directories.Size (Name);
			Ada.Processes.Append (Command, "-N");
			declare
				Dummy_File : Ada.Streams.Stream_IO.File_Type;
			begin
				Ada.Streams.Stream_IO.Create (
					Dummy_File,
					Name => Object.Temporary_Name.Constant_Reference);
				Ada.Streams.Stream_IO.Close (Dummy_File);
			end;
			Ada.Directories.Set_Modification_Time (
				Object.Temporary_Name.Constant_Reference,
				Object.Saved_Time);
		else
			Object.Saved_Time := Calendar.Null_Time;
			Object.Saved_Size := 0;
			if Ada.Directories.Exists (Object.Temporary_Name.Constant_Reference) then
				Ada.Debug.Put (Object.Temporary_Name.Constant_Reference);
				Ada.Directories.Delete_File (Object.Temporary_Name.Constant_Reference);
			end if;
		end if;
		Ada.Processes.Append (Command, "-nv"); -- no verbose
		Ada.Processes.Append (Command, "-nd"); -- no make directories
		if URI'Length >= 8 and then URI (URI'First .. URI'First + 7) = "https://" then
			Ada.Processes.Append (Command, "--no-check-certificate");
			Ada.Processes.Append (Command, URI);
		elsif URI'Length >= 7
			and then URI (URI'First .. URI'First + 6) = "feed://"
		then
			Ada.Processes.Append (
				Command,
				"http://" & URI (URI'First + 7 .. URI'Last));
		else
			Ada.Processes.Append (Command, URI);
		end if;
		Ada.Debug.Put (Object.Temporary_Name.Constant_Reference);
		Ada.Debug.Put (Ada.Processes.Image (Command));
		Ada.Processes.Create (
			Object.Wget,
			Command,
			Directory => Object.Temporary_Directory.Constant_Reference);
	end Start;
	
	procedure Wait (
		Object : in out Connection;
		Modified : out Boolean;
		Overwrite_Now : in Boolean := True)
	is
		Status : Ada.Command_Line.Exit_Status;
		New_Size : Ada.Directories.File_Size;
	begin
		Ada.Processes.Wait (Object.Wget, Status);
		if Status = 127 then
			Ada.Debug.Put (Ada.Environment_Variables.Value ("PATH"));
			raise Program_Error with "Wget is not installed.";
		elsif Status /= 0 then
			raise Not_Available;
		end if;
		Ada.Debug.Put (Object.Temporary_Name.Constant_Reference);
		Modified :=
			Ada.Directories.Modification_Time (Object.Temporary_Name.Constant_Reference) >
			Object.Saved_Time;
		if Modified then
			declare
				New_Temporary_Name : aliased Ada.Strings.Unbounded.Unbounded_String :=
					+Ada.Directories.Temporary.Create_Temporary_File;
			begin
				Rename_Or_Extract (
					Object.Temporary_Name.Constant_Reference,
					New_Temporary_Name.Constant_Reference);
				Object.Temporary_Name := New_Temporary_Name;
			end;
			New_Size := Ada.Directories.Size (Object.Temporary_Name.Constant_Reference);
			Ada.Debug.Put (
				Ada.Directories.File_Size'Image (Object.Saved_Size) & " -> "
					& Ada.Directories.File_Size'Image (New_Size));
			Modified := New_Size /= Object.Saved_Size;
			if Modified and then Overwrite_Now then
				Ada.Directories.Rename (
					Object.Temporary_Name.Constant_Reference,
					Object.Name.Constant_Reference,
					Overwrite => True);
			end if;
		end if;
		if not Modified then
			Ada.Directories.Delete_File (Object.Temporary_Name.Constant_Reference);
		end if;
		begin
			Ada.Directories.Delete_Directory (
				Object.Temporary_Directory.Constant_Reference);
		exception
			when E : Ada.Directories.Use_Error =>
				Ada.Debug.Put ("on deleting " & Object.Temporary_Directory.Constant_Reference);
				Ada.Debug.Put (Ada.Exceptions.Exception_Information (E));
		end;
	end Wait;
	
	function Temporary_Name (Object : Connection) return String is
	begin
		return Ada.Strings.Unbounded.To_String (Object.Temporary_Name);
	end Temporary_Name;
	
end Birdeyes.HTTP;
