with Ada.Calendar.Time_Zones;
with Ada.Formatting;
with Ada.Strings.Functions;
with Ada.Strings.Unbounded;
with Birdeyes.HTTP;
package body Birdeyes.LIRS is
	use type Ada.Calendar.Time;
	use type Ada.Strings.Unbounded.Unbounded_String;
	
	UNIX_Time_Base : constant Ada.Calendar.Time :=
		Ada.Calendar.Time_Of (1970, 1, 1);
	
	function To_UNIX_Time (Time : Ada.Calendar.Time) return Long_Long_Integer is
	begin
		return Long_Long_Integer (Time - UNIX_Time_Base);
	end To_UNIX_Time;
	
	function Image is new Ada.Formatting.Integer_Image (
		Long_Long_Integer,
		Signs => Ada.Formatting.Triming_Sign_Marks);
	
	LIRS_Signature : constant String := "LIRS";
	
	Connect_Count : constant := 8; -- connection count at same time
	
	-- implementation
	
	procedure Import (
		Syndication_Manager : in out Syndications.Management.Manager;
		File : in Ada.Text_IO.File_Type;
		Time : in Ada.Calendar.Time;
		Success_Item : out URI.Lists.List;
		Failure_Item : out URI.Lists.List) is
	begin
		while not Ada.Text_IO.End_Of_File (File) loop
			declare
				Imported :
					array (1 .. Connect_Count) of aliased Ada.Strings.Unbounded.Unbounded_String;
				Connections : array (Imported'Range) of HTTP.Connection;
				Last : Natural := 0;
			begin
				for I in Imported'Range loop
					exit when Ada.Text_IO.End_Of_File (File);
					declare
						Line : constant String := Ada.Text_IO.Get_Line (File);
						F : Positive;
						L : Natural;
					begin
						-- LIRS signature
						F := Line'First;
						L := Ada.Strings.Functions.Index_Element_Forward (Line, ',');
						if Line (F .. L - 1) /= LIRS_Signature then
							raise Data_Error with Line;
						end if;
						-- last-modified time (seconds, from 1970-01-01 GMT)
						F := L + 1;
						L := Ada.Strings.Functions.Index_Element_Forward (Line (F .. Line'Last), ',');
						-- last-detected time (seconds, from 1970-01-01 GMT)
						F := L + 1;
						L := Ada.Strings.Functions.Index_Element_Forward (Line (F .. Line'Last), ',');
						-- time zone (seconds)
						F := L + 1;
						L := Ada.Strings.Functions.Index_Element_Forward (Line (F .. Line'Last), ',');
						-- content-length
						F := L + 1;
						L := Ada.Strings.Functions.Index_Element_Forward (Line (F .. Line'Last), ',');
						-- URI
						F := L + 1;
						L := Ada.Strings.Functions.Index_Element_Forward (Line (F .. Line'Last), ',');
						Imported (I) := +URI.Normalize (Line (F .. L - 1));
						-- title
						-- author
						-- URI of agent
						-- extension (option)
					end;
					Ada.Debug.Put (Imported (I).Constant_Reference);
					Syndications.Management.Start_Update (
						Syndication_Manager,
						Imported (I).Constant_Reference,
						Connections (I));
					Last := I;
				end loop;
				for I in Imported'First .. Last loop
					declare
						Info : Syndications.Syndication_Info;
						Articles : Syndications.Article_Lists.List;
					begin
						Syndications.Management.Wait_Update (
							Syndication_Manager,
							Imported (I).Constant_Reference,
							Connections (I),
							Time,
							Info,
							Articles);
						URI.Lists.Append (Success_Item, Imported (I));
					exception
						when Syndications.Management.Not_Available =>
							URI.Lists.Append (Failure_Item, Imported (I));
					end;
				end loop;
			end;
		end loop;
	end Import;
	
	procedure Export (
		Syndication_Manager : in out Syndications.Management.Manager;
		File : in Ada.Text_IO.File_Type;
		Item : in URI.Lists.List;
		Agent_URI : in String)
	is
		-- birdeyes does not save time zones of each syndication
		Time_Zone : constant String :=
			Image (Long_Long_Integer (Ada.Calendar.Time_Zones.UTC_Time_Offset) * 60);
	begin
		for I in Item.Iterate loop
			declare
				URI : String renames Item.Constant_Reference (I).Constant_Reference;
				Syndication_Info : Syndications.Syndication_Info;
			begin
				Syndications.Management.Query (
					Syndication_Manager,
					URI,
					Syndication_Info);
				-- LIRS signature
				Ada.Text_IO.Put (LIRS_Signature & ",");
				-- last-modified time (seconds, from 1970-01-01 GMT)
				Ada.Text_IO.Put (
					Image (To_UNIX_Time (Syndications.Last_Modified_Time (Syndication_Info)))
						& ",");
				-- last-detected time (seconds, from 1970-01-01 GMT)
				Ada.Text_IO.Put (
					Image (To_UNIX_Time (Syndication_Info.Last_Avail_Time)) & ",");
				-- time zone (seconds)
				Ada.Text_IO.Put (Time_Zone & ",");
				-- content-length
				Ada.Text_IO.Put (
					Image (Long_Long_Integer (Syndication_Info.Content_Length)) & ",");
				-- URI
				Ada.Text_IO.Put (URI & ",");
				-- title
				Ada.Text_IO.Put (Syndication_Info.Title.Constant_Reference & ",");
				-- author
				Ada.Text_IO.Put (Syndication_Info.Author.Constant_Reference & ",");
				-- URI of agent
				Ada.Text_IO.Put (Agent_URI & ",");
				-- extension (option)
				Ada.Text_IO.Put (",");
				-- line feed
				Ada.Text_IO.New_Line (File);
			end;
		end loop;
	end Export;
	
end Birdeyes.LIRS;
