with Ada.Calendar;
with Ada.Directories;
with Ada.Exception_Identification.From_Here;
with Ada.Streams.Stream_IO;
package body Birdeyes.Pages is
	use type Ada.Calendar.Time;
	use type Ada.Strings.Unbounded.Unbounded_String;
	
	function Parameters_To_Index_Page
		return Web.Query_Strings is
	begin
		return Web.String_Maps.Empty_Map;
	end Parameters_To_Index_Page;
	
	function Parameters_To_View_Page (
		View_User_Id : String)
		return Web.Query_Strings is
	begin
		return Result : Web.Query_Strings do
			Web.Include (Result, "view", View_User_Id);
		end return;
	end Parameters_To_View_Page;
	
	function Parameters_To_RSS_Page (
		View_User_Id : String)
		return Web.Query_Strings is
	begin
		return Result : Web.Query_Strings do
			Web.Include (Result, "view", View_User_Id);
			Web.Include (Result, "rss", "2.0");
		end return;
	end Parameters_To_RSS_Page;
	
	function Where (
		Query_Strings : Web.Query_Strings)
		return Page is
	begin
		if Query_Strings.Contains ("view") then
			if Query_Strings.Contains ("rss") then
				return RSS;
			else
				return View;
			end if;
		else
			return Index;
		end if;
	end Where;
	
	function Get_View_User_Id (
		Query_Strings : Web.Query_Strings;
		User_Manager : Users.Management.Manager)
		return String
	is
		Result : constant String := Web.Element (Query_Strings, "view");
	begin
		if Result'Length /= 0 then
			return Users.Management.Registered_Id (User_Manager, Result);
		else
			return "";
		end if;
	end Get_View_User_Id;
	
	function Get_User_Id (
		Cookie : Web.Cookie;
		User_Manager : Users.Management.Manager)
		return String
	is
		Result : constant String := Web.Element (Cookie, "be_id");
	begin
		if Result'Length /= 0 then
			return Users.Management.Registered_Id (User_Manager, Result);
		else
			return "";
		end if;
	end Get_User_Id;
	
	function Get_User_Password (
		Cookie : Web.Cookie)
		return String is
	begin
		return Web.Element (Cookie, "be_password");
	end Get_User_Password;
	
	procedure Set_User (
		Cookie : in out Web.Cookie;
		New_User_Id: in String;
		New_User_Password : in String) is
	begin
		Web.Include (Cookie, "be_id", New_User_Id);
		Web.Include (Cookie, "be_password", New_User_Password);
	end Set_User;
	
	function Get_Command (
		Inputs : Web.Query_Strings)
		return String is
	begin
		return Web.Element (Inputs, "cmd");
	end Get_Command;
	
	function Get_New_User_Id (
		Inputs : Web.Query_Strings)
		return String is
	begin
		return Web.Element (Inputs, "be_id");
	end Get_New_User_Id;
	
	function Get_New_User_Password (
		Inputs : Web.Query_Strings)
		return String is
	begin
		return Web.Element (Inputs, "be_password");
	end Get_New_User_Password;
	
	function Get_New_User_Confirmation_Password (
		Inputs : Web.Query_Strings)
		return String is
	begin
		return Web.Element (Inputs, "be_password2");
	end Get_New_User_Confirmation_Password;
	
	function Get_URI (
		Inputs : Web.Query_Strings)
		return String is
	begin
		return Web.Element (Inputs, "uri");
	end Get_URI;
	
	function Get_Showing (
		Inputs : Web.Query_Strings)
		return Users.Showing_Mode is
	begin
		return Users.Showing_Mode'Value (Web.Element (Inputs, "showing"));
	end Get_Showing;
	
	function Get_Content (
		Inputs : Web.Query_Strings)
		return String is
	begin
		return Web.Element (Inputs, "content");
	end Get_Content;
	
	procedure Get_Message (
		Cookie : in out Web.Cookie;
		Message : out Ada.Strings.Unbounded.Unbounded_String) is
	begin
		Message := +Web.Element (Cookie, "message");
		Web.Include (Cookie, "message", "");
	end Get_Message;
	
	function Has_Message (Cookie : Web.Cookie) return Boolean is
	begin
		return String'(Web.Element (Cookie, "message"))'Length /= 0;
	end Has_Message;
	
	procedure Set_Message (
		Cookie : in out Web.Cookie;
		Message : in String) is
	begin
		Web.Include (Cookie, "message", Message);
	end Set_Message;
	
	function Read (
		Template_Source : in String;
		Template_Cache : in String := "")
		return Web.Producers.Template
	is
		File : Ada.Streams.Stream_IO.File_Type :=
			Ada.Streams.Stream_IO.Open (
				Ada.Streams.Stream_IO.In_File,
				Name => Template_Source);
	begin
		return Template : Web.Producers.Template :=
			Web.Producers.Read (
				Ada.Streams.Stream_IO.Stream (File),
				Ada.Streams.Stream_IO.Size(File),
				Parsing => False)
		do
			if Template_Cache'Length /= 0 then
				if Ada.Directories.Exists (Template_Cache)
					and then Ada.Directories.Modification_Time (Template_Cache) >
						Ada.Directories.Modification_Time (Template_Source)
				then
					-- read parsed-structure from cache file
					declare
						Cache_File : Ada.Streams.Stream_IO.File_Type;
					begin
						Ada.Streams.Stream_IO.Open (
							Cache_File,
							Ada.Streams.Stream_IO.In_File,
							Name => Template_Cache);
						Web.Producers.Read_Parsed_Information (
							Ada.Streams.Stream_IO.Stream (Cache_File),
							Template);
						Ada.Streams.Stream_IO.Close (Cache_File);
					end;
				else
					Web.Producers.Parse (Template);
					-- save parsed-structure to cache file
					declare
						Cache_File : Ada.Streams.Stream_IO.File_Type;
					begin
						Ada.Streams.Stream_IO.Create (Cache_File, Name => Template_Cache);
						Web.Producers.Write_Parsed_Information (
							Ada.Streams.Stream_IO.Stream (Cache_File),
							Template);
						Ada.Streams.Stream_IO.Close (Cache_File);
					end;
				end if;
			else
				Web.Producers.Parse (Template);
			end if;
			Ada.Streams.Stream_IO.Close(File);
		end return;
	end Read;
	
	procedure Raise_Unknown_Tag (
		Tag : in String;
		File : in String := Ada.Debug.File;
		Line : in Integer := Ada.Debug.Line) is
	begin
		Ada.Exception_Identification.From_Here.Raise_Exception (
			Web.Producers.Data_Error'Identity,
			File => File,
			Line => Line,
			Message => "Invalid template """ & Tag & """");
	end Raise_Unknown_Tag;
	
end Birdeyes.Pages;
