with Ada.Calendar;
--with Ada.Calendar.Formatting;
with Ada.Containers.Generic_Array_Sort;
with Ada.Directories;
with Ada.Hierarchical_File_Names;
with Ada.Streams.Stream_IO;
with Crypto.SHA1;
with Birdeyes.Syndications.Serialization;
with Birdeyes.YAML_IO;
package body Birdeyes.Syndications.Management is
	use type Ada.Strings.Unbounded.Unbounded_String;
	
	procedure Save is new YAML_IO.Save (
		Serialization.Tag,
		Syndication_Info,
		Serialization.IO);
	procedure Load is new YAML_IO.Load (
		Serialization.Tag,
		Syndication_Info,
		Serialization.IO);
	
	procedure Read_RSS (
		Name : in String;
		Info : in out Syndication_Info;
		Articles : out Article_Lists.List)
	is
		File : Ada.Streams.Stream_IO.File_Type :=
			Ada.Streams.Stream_IO.Open (Ada.Streams.Stream_IO.In_File, Name => Name);
	begin
		Read_RSS (Ada.Streams.Stream_IO.Stream (File), Info, Articles);
		Ada.Streams.Stream_IO.Close (File);
	end Read_RSS;
	
	function Name (Object : Manager; URI : String) return String is
		Hash : Crypto.SHA1.Fingerprint;
		SHA1 : Crypto.SHA1.Context := Crypto.SHA1.Initial;
	begin
		Crypto.SHA1.Update (SHA1, URI);
		Crypto.SHA1.Final (SHA1, Hash);
		return Ada.Hierarchical_File_Names.Compose (
			Object.Path.all,
			Crypto.SHA1.Image (Hash));
	end Name;
	
	Max_Span : constant Duration := 24.0 * 60.0 * 60.0;
	
	function Key (Object : Manager; URI : String) return Ada.Calendar.Time is
		Info : Syndication_Info;
		Num : Natural := 0;
		Sum : Duration := 0.0;
		Span : Duration := 0.0;
		Result : Ada.Calendar.Time;
	begin
		Load (Name (Object, URI), Info);
		if Info.History.Length >= 2 then
			declare
				First : constant History_Lists.Cursor := Info.History.First;
				Previous : Ada.Calendar.Time :=
					Info.History.Constant_Reference (First).Modified_Time;
			begin
				for I in
					Info.History.Iterate (History_Lists.Next (First), Info.History.Last)
				loop
					declare
						Current : constant Ada.Calendar.Time :=
							Info.History.Constant_Reference (I).Modified_Time;
					begin
						pragma Assert (Current >= Previous);
						Num := Num + 1;
						Sum := Sum + (Current - Previous);
						Previous := Current;
					end;
				end loop;
			end;
			if Num /= 0 then
				Span := Sum / Num;
			end if;
		end if;
		pragma Assert (Info.Last_Tried_Time >= Info.Last_Avail_Time);
		Result := Info.Last_Tried_Time + Duration'Min (
			Max_Span,
			Duration'Max (Span, Info.Last_Tried_Time - Info.Last_Avail_Time));
		return Result;
	end Key;
	
	RSS_Extension : constant String := "rss";
	
	-- implementation
	
	function Create (Path : Static_String_Access) return Manager is
	begin
		return (Path => Path);
	end Create;
	
	procedure Query (
		Object : in out Manager;
		URI : in String;
		Info : out Syndication_Info)
	is
		Name : constant String := Management.Name (Object, URI);
	begin
		Load (Name, Info);
	end Query;
	
	procedure Query (
		Object : in out Manager;
		URI : in String;
		Info : out Syndication_Info;
		Articles : out Article_Lists.List)
	is
		Name : constant String := Management.Name (Object, URI);
		RSS_Name : constant String :=
			Ada.Hierarchical_File_Names.Compose (
				Relative_Name => Name,
				Extension => RSS_Extension);
	begin
		Load (Name, Info);
		begin
			Read_RSS (RSS_Name, Info, Articles);
		exception
			when Ada.Streams.Stream_IO.Name_Error => -- the saved rss is missing
				Article_Lists.Clear (Articles);
		end;
	end Query;
	
	procedure Start_Update (
		Object : in out Manager;
		URI : in String;
		Connection : in out HTTP.Connection)
	is
		Name : constant String := Management.Name (Object, URI);
		RSS_Name : constant String :=
			Ada.Hierarchical_File_Names.Compose (
				Relative_Name => Name,
				Extension => RSS_Extension);
	begin
		HTTP.Start (Connection, URI => URI, Name => RSS_Name);
	end Start_Update;
	
	procedure Wait_Update (
		Object : in out Manager;
		URI : in String;
		Connection : in out HTTP.Connection;
		Time : in Ada.Calendar.Time;
		Info : out Syndication_Info;
		Articles : out Article_Lists.List)
	is
		Name : constant String := Management.Name (Object, URI);
		RSS_Name : constant String :=
			Ada.Hierarchical_File_Names.Compose (
				Relative_Name => Name,
				Extension => RSS_Extension);
		First : constant Boolean := not Ada.Directories.Exists (Name);
		Avail : Boolean;
		Modified : Boolean;
		Read : Boolean;
	begin
		if First then
			Info.URI := +URI; -- before calling Read
			Info.Last_Avail_Time := Calendar.Null_Time;
			Info.Last_Detected_Time := Calendar.Null_Time;
		else
			Load (Name, Info);
		end if;
		Info.Last_Tried_Time := Time;
		begin
			HTTP.Wait (Connection, Modified, Overwrite_Now => False);
			Avail := True;
		exception
			when Not_Available =>
				if First then
					raise;
				else
					Avail := False;
					Modified := False;
				end if;
		end;
		pragma Assert (Avail >= Modified);
		Read := False;
		if Avail then
			if Modified then
				declare
					Temporary_Name : constant String := HTTP.Temporary_Name (Connection);
				begin
					Ada.Debug.Put (Temporary_Name);
					-- read
					Read_RSS (Temporary_Name, Info, Articles);
					Read := True;
					-- save
					Ada.Directories.Rename (Temporary_Name, RSS_Name, Overwrite => True);
					-- update content-length
					Info.Content_Length := Ada.Streams.Stream_Element_Offset (
						Ada.Directories.Size (RSS_Name));
					-- update timestamps
					Info.Last_Detected_Time := Time;
					Add_History (Info.History, Ada.Directories.Modification_Time (RSS_Name));
				exception
					when Not_RSS =>
						Ada.Directories.Delete_File (Temporary_Name);
						if First then
							raise Not_Available;
						end if;
				end;
			end if;
			Info.Last_Avail_Time := Time;
		end if;
		Save (Name, Info);
		if not Read then
			begin
				Read_RSS (RSS_Name, Info, Articles);
			exception
				when Ada.Streams.Stream_IO.Name_Error => -- the saved rss is missing
					Article_Lists.Clear (Articles);
			end;
		end if;
	end Wait_Update;
	
	procedure Get (
		Object : in out Manager;
		URI : in String;
		Time : in Ada.Calendar.Time;
		Info : out Syndication_Info;
		Articles : out Article_Lists.List)
	is
		Connection : HTTP.Connection;
	begin
		Ada.Directories.Create_Path (Object.Path.all);
		Start_Update (Object, URI, Connection);
		Wait_Update (Object, URI, Connection, Time, Info, Articles);
	end Get;
	
	function Select_For_Update (
		Object : Manager;
		URI_List : URI.Lists.List;
		Num : Natural := Connect_Count)
		return URI.Lists.List
	is
		Length : Natural := URI_List.Length;
		type T is record
			Key : Ada.Calendar.Time;
			URI : aliased Ada.Strings.Unbounded.Unbounded_String;
		end record;
		function "<" (Left, Right : T) return Boolean is
		begin
			return Left.Key < Right.Key;
		end "<";
		type T_Array is array (Positive range <>) of T;
		procedure Sort is
			new Ada.Containers.Generic_Array_Sort (Positive, T, T_Array);
		Ordered_List : T_Array (1 .. Length);
	begin
		declare
			Index : Positive := Ordered_List'First;
		begin
			for I in URI_List.Iterate loop
				Ordered_List (Index).URI := URI_List.Constant_Reference (I);
				Ordered_List (Index).Key :=
					Key (Object, Ordered_List (Index).URI.Constant_Reference);
				Index := Index + 1;
			end loop;
		end;
		Sort (Ordered_List);
		return Result : URI.Lists.List do
			for I in 1 .. Natural'Min (Num, Length) loop
				URI.Lists.Append (Result, Ordered_List (I).URI);
			end loop;
		end return;
	end Select_For_Update;
	
	procedure Get (
		Object : in out Manager;
		URI_List : in URI.Lists.List;
		Update_URI_List : in URI.Lists.List;
		Time : in Ada.Calendar.Time;
		Result : out Syndication_Info_And_Articles_Lists.List) is
	begin
		Ada.Directories.Create_Path (Object.Path.all);
		declare
			Position : URI.Lists.Cursor := Update_URI_List.First;
		begin
			while URI.Lists.Has_Element (Position) loop
				declare
					URIs :
						array (1 .. Connect_Count) of aliased Ada.Strings.Unbounded.Unbounded_String;
					Connections : array (URIs'Range) of HTTP.Connection;
					Last : Natural := 0;
				begin
					for I in Connections'Range loop
						exit when not URI.Lists.Has_Element (Position);
						URIs (I) := Update_URI_List.Constant_Reference (Position);
						Position := URI.Lists.Next (Position);
						Ada.Debug.Put (URIs (I).Constant_Reference);
						Syndications.Management.Start_Update (
							Object,
							URIs (I).Constant_Reference,
							Connections (I));
						Last := I;
					end loop;
					for I in URIs'First .. Last loop
						declare
							Syndication : Syndication_Info_And_Articles;
						begin
							Syndications.Management.Wait_Update (
								Object,
								URIs (I).Constant_Reference,
								Connections (I),
								Time,
								Syndication.Info,
								Syndication.Articles);
							Syndication_Info_And_Articles_Lists.Append (Result, Syndication);
						exception
							when Syndications.Management.Not_Available =>
								declare
									Syndication : Syndication_Info_And_Articles;
								begin
									Query (
										Object,
										URIs (I).Constant_Reference,
										Syndication.Info,
										Syndication.Articles);
									Syndication_Info_And_Articles_Lists.Append (Result, Syndication);
								end;
						end;
					end loop;
				end;
			end loop;
		end;
		for I in URI_List.Iterate loop
			declare
				URI : Ada.Strings.Unbounded.Unbounded_String
					renames URI_List.Constant_Reference (I);
			begin
				if not Update_URI_List.Contains (URI) then
					declare
						Syndication : Syndication_Info_And_Articles;
					begin
						Query (
							Object,
							URI.Constant_Reference,
							Syndication.Info,
							Syndication.Articles);
						Syndication_Info_And_Articles_Lists.Append (Result, Syndication);
					end;
				end if;
			end;
		end loop;
	end Get;
	
end Birdeyes.Syndications.Management;
