with Ada.Calendar.Formatting;
with Ada.Calendar.Time_Zones;
with Ada.Exceptions;
with XML.Streams;
with Web;
package body Birdeyes.Syndications is
	use type Ada.Strings.Unbounded.Unbounded_String;
	
	function DC_Date_Value (Image : String) return Ada.Calendar.Time is
	begin
		if Image'Length = 10 then -- "yyyy-mm-dd"
			return Ada.Calendar.Formatting.Value (Image & " 00:00:00");
		elsif Image'Length = 25 and then Image (Image'First + 10) = 'T' then
			-- yyyy-mm-ddThh-mm-ss+hh:mm
			declare
				Image_2 : String := Image;
				Offset : Ada.Calendar.Time_Zones.Time_Offset;
			begin
				Image_2 (Image_2'First + 10) := ' ';
				Offset :=
					Ada.Calendar.Formatting.Value (
						Image_2 (Image_2'First + 19 .. Image_2'Last));
				return Ada.Calendar.Formatting.Value (
					Image_2 (Image_2'First .. Image_2'First + 18), Time_Zone => Offset);
			end;
		else
			Ada.Debug.Put (Image);
			return Calendar.Null_Time;
		end if;
	exception
		when E : Constraint_Error =>
			Ada.Debug.Put (Image);
			Ada.Debug.Put (Ada.Exceptions.Exception_Information (E));
			return Calendar.Null_Time;
	end DC_Date_Value;
	
	-- implementation
	
	function Image (Value : Syndication_Version) return String is
	begin
		case Value is
			when Unknown =>
				return "unknown";
			when RSS_1 =>
				return "RSS 1.0";
			when RSS_2 =>
				return "RSS 2.0";
			when Atom =>
				return "Atom";
		end case;
	end Image;
	
	procedure Add_History (
		Container : in out History_Lists.List;
		Modified_Time : in Ada.Calendar.Time) is
	begin
		while not Container.Is_Empty
			and then Modified_Time
					- Container.Constant_Reference (Container.First).Modified_Time >
				Modified_Time_Limit
		loop
			History_Lists.Delete_First (Container);
		end loop;
		History_Lists.Append (Container, (Modified_Time => Modified_Time));
	end Add_History;
	
	function Last_Modified_Time (Info : Syndication_Info)
		return Ada.Calendar.Time is
	begin
		pragma Assert (not Info.History.Is_Empty);
		return Info.History.Constant_Reference (Info.History.Last).Modified_Time;
	end Last_Modified_Time;
	
	function Find_Latest (Container : Article_Lists.List; Time : Ada.Calendar.Time)
		return Article_Lists.Cursor is
	begin
		return Result : Article_Lists.Cursor := Container.First do
			if Container.Length >= 2 then
				for I in Container.Iterate (Article_Lists.Next (Result), Container.Last) loop
					if Container.Constant_Reference (I).Time >
						Container.Constant_Reference (Result).Time
						or else (
							Container.Constant_Reference (I).Time <= Time
							and then Container.Constant_Reference (Result).Time > Time)
					then
						Result := I;
					end if;
				end loop;
			end if;
		end return;
	end Find_Latest;
	
	procedure Read_RSS (
		Stream : not null access Ada.Streams.Root_Stream_Type'Class;
		Info : in out Syndication_Info;
		Articles : out Article_Lists.List) is
	begin
		declare
			Reader : XML.Reader :=
				XML.Streams.Create (Stream, URI => Info.URI.Constant_Reference);
			type T is (Root, RSS, Channel, Link, Title, Author, Item,
				Item_Link, Item_Title, Item_Author,
				Item_Description, Item_Content, Item_Source, Item_Time,
				Unknown);
			S : array (1 .. 8) of T := (1 => Root, others => Unknown);
			Top : Natural := S'First;
			procedure Push (N : in T) is
			begin
				Top := Top + 1;
				S (Top) := N;
			end Push;
			procedure Pop is
			begin
				Top := Top - 1;
			end Pop;
		begin
			Info.Version := Unknown;
			loop
				if S (Top) = Unknown then
					XML.Read_Until_Element_End (Reader);
					Pop;
				else
					declare
						Parsing_Entry : XML.Parsing_Entry_Type;
					begin
						begin
							XML.Read (Reader, Parsing_Entry);
						exception
							when XML.Use_Error => -- ???
								Ada.Debug.Put ("from Use_Error to Data_Error");
								raise XML.Data_Error;
						end;
						declare
							Event : XML.Event renames XML.Value (Parsing_Entry);
						begin
							case S (Top) is
								when Root =>
									case Event.Event_Type is
										when XML.Element_Start =>
											if Event.Name.all = "rdf:RDF" then
												Info.Version := RSS_1;
												Push (RSS);
											elsif Event.Name.all = "rss" then
												Info.Version := RSS_2;
												Push (RSS);
											elsif Event.Name.all = "feed" then
												Info.Version := Atom;
												raise XML.Data_Error with
													"ATOM is not supported.";
											else
												Ada.Debug.Put (Event.Name.all);
												raise XML.Data_Error;
											end if;
										when XML.Processing_Instruction =>
											null;
										when XML.Comment =>
											null;
										when XML.No_Event =>
											Pop;
										when others =>
											Ada.Debug.Put (
												XML.Event_Type'Image (Event.Event_Type));
											raise XML.Data_Error;
									end case;
								when RSS =>
									case Event.Event_Type is
										when XML.Attribute =>
											null; -- skip xmlns
										when XML.Element_Start =>
											if Event.Name.all = "channel" then
												Push (Channel);
											elsif Info.Version = RSS_1
												and then Event.Name.all = "item"
											then
												Article_Lists.Append (
													Articles,
													(others => <>));
												Push (Item);
											else
												Push (Unknown);
											end if;
										when XML.Significant_Whitespace =>
											null;
										when XML.Element_End =>
											Pop;
										when others =>
											Ada.Debug.Put (
												XML.Event_Type'Image (Event.Event_Type));
											raise XML.Data_Error;
									end case;
								when Channel =>
									case Event.Event_Type is
										when XML.Attribute =>
											if Event.Name.all = "rdf:about" then
												Info.Link := +Event.Value.all;
											end if;
										when XML.Element_Start =>
											if Event.Name.all = "link" then
												Push (Link);
											elsif Event.Name.all = "title" then
												Push (Title);
											elsif Event.Name.all = "dc:creator" then
												Push (Author);
											elsif Event.Name.all = "managingEditor"
												and then Info.Author.Is_Null
											then
												Push (Author);
											elsif Info.Version = RSS_2
												and then Event.Name.all = "item"
											then
												Article_Lists.Append (
													Articles,
													(others => <>));
												Push (Item);
											else
												Push (Unknown);
											end if;
										when XML.Significant_Whitespace =>
											null;
										when XML.Element_End =>
											Pop;
										when others =>
											Ada.Debug.Put (
												XML.Event_Type'Image (Event.Event_Type));
											raise XML.Data_Error;
									end case;
								when Link =>
									case Event.Event_Type is
										when XML.Text =>
											Info.Link := +Event.Content.all;
										when XML.Element_End =>
											Pop;
										when others =>
											Ada.Debug.Put (
												XML.Event_Type'Image (Event.Event_Type));
											raise XML.Data_Error;
									end case;
								when Title =>
									case Event.Event_Type is
										when XML.Text | XML.CDATA =>
											Info.Title := +Event.Content.all;
										when XML.Significant_Whitespace =>
											null;
										when XML.Element_End =>
											Pop;
										when others =>
											Ada.Debug.Put (
												XML.Event_Type'Image (Event.Event_Type));
											raise XML.Data_Error;
									end case;
								when Author =>
									case Event.Event_Type is
										when XML.Text | XML.CDATA =>
											Info.Author := +Event.Content.all;
										when XML.Significant_Whitespace =>
											null;
										when XML.Element_End =>
											Pop;
										when others =>
											Ada.Debug.Put (
												XML.Event_Type'Image (Event.Event_Type));
											raise XML.Data_Error;
									end case;
								when Item =>
									case Event.Event_Type is
										when XML.Attribute =>
											if Event.Name.all = "rdf:about" then
												declare
													Last_Position : constant Article_Lists.Cursor :=
														Articles.Last;
												begin
													Articles.Reference (Last_Position).Link :=
														+Event.Value.all;
												end;
											end if;
										when XML.Element_Start =>
											if Event.Name.all = "title" then
												Push (Item_Title);
											elsif Event.Name.all = "link" then
												Push (Item_Link);
											elsif Event.Name.all = "author" then
												Push (Item_Author);
											elsif Event.Name.all = "dc:creator" then
												Push (Item_Author);
											elsif Event.Name.all = "source" then
												Push (Item_Source);
											elsif Event.Name.all = "pubDate" then
												Push (Item_Time);
											elsif Event.Name.all = "dc:date" then
												Push (Item_Time);
											elsif Event.Name.all = "description" then
												Push (Item_Description);
											elsif Event.Name.all = "content:encoded" then
												Push (Item_Content);
											else
												Push (Unknown);
											end if;
										when XML.Significant_Whitespace =>
											null;
										when XML.Element_End =>
											Pop;
										when others =>
											Ada.Debug.Put (
												XML.Event_Type'Image (Event.Event_Type));
											raise XML.Data_Error;
									end case;
								when Item_Link =>
									case Event.Event_Type is
										when XML.Text =>
											declare
												Last_Position : constant Article_Lists.Cursor :=
													Articles.Last;
											begin
												Articles.Reference (Last_Position).Link :=
													+Event.Content.all;
											end;
										when XML.Element_End =>
											Pop;
										when others =>
											Ada.Debug.Put (
												XML.Event_Type'Image (Event.Event_Type));
											raise XML.Data_Error;
									end case;
								when Item_Title =>
									case Event.Event_Type is
										when XML.Text | XML.CDATA =>
											declare
												Last_Position : constant Article_Lists.Cursor :=
													Articles.Last;
											begin
												Articles.Reference (Last_Position).Title :=
													+Event.Content.all;
											end;
										when XML.Significant_Whitespace =>
											null;
										when XML.Element_End =>
											Pop;
										when others =>
											Ada.Debug.Put (
												XML.Event_Type'Image (Event.Event_Type));
											raise XML.Data_Error;
									end case;
								when Item_Author =>
									case Event.Event_Type is
										when XML.Text | XML.CDATA =>
											declare
												Last_Position : constant Article_Lists.Cursor :=
													Articles.Last;
												Last_Article : Article
													renames Articles.Reference (
														Last_Position);
											begin
												Last_Article.Author := +Event.Content.all;
												if Info.Author.Is_Null then
													Info.Author := Last_Article.Author;
												end if;
											end;
										when XML.Significant_Whitespace =>
											null;
										when XML.Element_End =>
											Pop;
										when others =>
											Ada.Debug.Put (
												XML.Event_Type'Image (Event.Event_Type));
											raise XML.Data_Error;
									end case;
								when Item_Source =>
									case Event.Event_Type is
										when XML.Attribute =>
											if Event.Name.all = "url" then
												declare
													Last_Position : constant Article_Lists.Cursor :=
														Articles.Last;
												begin
													Articles.Reference (Last_Position).Source_URI :=
														+Event.Value.all;
												end;
											end if;
										when XML.Text | XML.CDATA =>
											declare
												Last_Position : constant Article_Lists.Cursor :=
													Articles.Last;
											begin
												Articles.Reference (Last_Position).Source_Title :=
													+Event.Content.all;
											end;
										when XML.Significant_Whitespace =>
											null;
										when XML.Element_End =>
											Pop;
										when others =>
											Ada.Debug.Put (
												XML.Event_Type'Image (Event.Event_Type));
											raise XML.Data_Error;
									end case;
								when Item_Time =>
									case Event.Event_Type is
										when XML.Text | XML.CDATA =>
											declare
												Last_Position : constant Article_Lists.Cursor :=
													Articles.Last;
												The_Time : Ada.Calendar.Time;
											begin
												begin
													The_Time :=
														Web.Value (Event.Content.all);
												exception
													when Constraint_Error =>
														The_Time :=
															DC_Date_Value (Event.Content.all);
												end;
												Articles.Reference (Last_Position).Time :=
													The_Time;
											end;
										when XML.Significant_Whitespace =>
											null;
										when XML.Element_End =>
											Pop;
										when others =>
											Ada.Debug.Put (
												XML.Event_Type'Image (Event.Event_Type));
											raise XML.Data_Error;
									end case;
								when Item_Description =>
									case Event.Event_Type is
										when XML.Text | XML.CDATA =>
											declare
												Last_Position : constant Article_Lists.Cursor :=
													Articles.Last;
												Last_Article : Article
													renames Articles.Reference (
														Last_Position);
											begin
												Last_Article.Description :=
													+Event.Content.all;
												if Last_Article.Content.Is_Null then
													Last_Article.Content :=
														Last_Article.Description;
												end if;
											end;
										when XML.Significant_Whitespace =>
											null;
										when XML.Element_End =>
											Pop;
										when others =>
											Ada.Debug.Put (
												XML.Event_Type'Image (Event.Event_Type));
											raise XML.Data_Error;
									end case;
								when Item_Content =>
									case Event.Event_Type is
										when XML.Text | XML.CDATA =>
											declare
												Last_Position : constant Article_Lists.Cursor :=
													Articles.Last;
												Last_Article : Article
													renames Articles.Reference (
														Last_Position);
											begin
												Last_Article.Content := +Event.Content.all;
												if Last_Article.Description.Is_Null then
													Last_Article.Description :=
														Last_Article.Content;
												end if;
											end;
										when XML.Significant_Whitespace =>
											null;
										when XML.Element_End =>
											Pop;
										when others =>
											Ada.Debug.Put (
												XML.Event_Type'Image (Event.Event_Type));
											raise XML.Data_Error;
									end case;
								when Unknown =>
									raise Program_Error;
							end case;
						end;
					end;
				end if;
				exit when Top < S'First;
			end loop;
		end;
	exception
		when E : XML.Data_Error =>
			Ada.Debug.Put ("on parsing " & Info.URI.Constant_Reference);
			Ada.Debug.Put (Ada.Exceptions.Exception_Information (E));
			raise Not_RSS;
		when E : XML.Status_Error =>
			Ada.Debug.Put ("* birdeyes (or libxml2) maybe have a bug???");
			Ada.Debug.Put ("on parsing " & Info.URI.Constant_Reference);
			Ada.Debug.Put (Ada.Exceptions.Exception_Information (E));
			raise Not_RSS;
	end Read_RSS;
	
end Birdeyes.Syndications;
