with Ada.Characters.ASCII.Handling;
with Ada.Directories.Information;
with Ada.Hierarchical_File_Names;
with Birdeyes.Users.Serialization;
with Birdeyes.YAML_IO;
package body Birdeyes.Users.Management is
	use type Ada.Strings.Unbounded.Unbounded_String;
	
	procedure Save is new YAML_IO.Save (
		Serialization.Tag,
		User_Info,
		Serialization.IO);
	procedure Load is new YAML_IO.Load (
		Serialization.Tag,
		User_Info,
		Serialization.IO);
	
	function Name (Object : Manager; Id : String) return String is
	begin
		return Ada.Hierarchical_File_Names.Compose (Object.Path.all, Id);
	end Name;
	
	-- implementation
	
	function Create (Path : Static_String_Access) return Manager is
	begin
		return (Path => Path);
	end Create;
	
	procedure Iterate (
		Object : in Manager;
		Process : not null access procedure (Id : in String)) is
	begin
		if Ada.Directories.Exists (Object.Path.all) then
			declare
				Search : aliased Ada.Directories.Search_Type;
			begin
				Ada.Directories.Start_Search (Search,
					Directory => Object.Path.all,
					Filter => (Ada.Directories.Ordinary_File => True, others => False));
				while Ada.Directories.More_Entries(Search) loop
					declare
						Directory_Entry : Ada.Directories.Directory_Entry_Type
							renames Ada.Directories.Look_Next_Entry (Search);
						Name : String
							renames Ada.Directories.Simple_Name (Directory_Entry);
					begin
						if Valid_Id_String (Name) then
							pragma Assert (
								not Ada.Directories.Information.Is_Symbolic_Link (
									Ada.Directories.Full_Name (Directory_Entry)));
							Process (Name);
						end if;
					end;
					Ada.Directories.Skip_Next_Entry (Search);
				end loop;
				Ada.Directories.End_Search (Search);
			end;
		end if;
	end Iterate;
	
	function Exists (Object : Manager; Id : String) return Boolean is
	begin
		if Ada.Directories.Exists (Name (Object, Id)) then
			return True;
		else
			declare
				Lower_Case_Id : constant String := Ada.Characters.ASCII.Handling.To_Lower (Id);
			begin
				return Id /= Lower_Case_Id
					and then Ada.Directories.Exists (Name (Object, Id));
			end;
		end if;
	end Exists;
	
	function Registered_Id (Object : Manager; Id : String) return String is
		Name : constant String := Management.Name (Object, Id);
	begin
		if Ada.Directories.Information.Is_Symbolic_Link (Name) then
			return Ada.Directories.Simple_Name (
				Ada.Directories.Information.Read_Symbolic_Link (Name));
		else
			return Id;
		end if;
	end Registered_Id;
	
	procedure Add (
		Object : in out Manager;
		Id : in String;
		Password : in Password_Digest;
		Remote_Addr : in String;
		Remote_Host : in String;
		Time : in Ada.Calendar.Time;
		Info : out User_Info) is
	begin
		if not Valid_Id_String (Id) or else Exists (Object, Id) then
			raise Name_Error;
		end if;
		Info := (
			Password => Password,
			Creation_Remote_Addr => +Remote_Addr,
			Creation_Remote_Host => +Remote_Host,
			Creation_Time => Time,
			Last_Remote_Addr => +Remote_Addr,
			Last_Remote_Host => +Remote_Host,
			Last_Time => Time,
			Watching => Watching_Syndications.Empty_List);
		declare
			Name : constant String := Management.Name (Object, Id);
		begin
			-- create the directory
			declare
				Dir : constant String :=
					Ada.Hierarchical_File_Names.Unchecked_Containing_Directory (Name);
			begin
				if Dir'Length /= 0 then
					Ada.Directories.Create_Path (Dir);
				end if;
			end;
			-- write the file
			Save (Name, Info);
			declare
				Lower_Case_Id : constant String := Ada.Characters.ASCII.Handling.To_Lower (Id);
			begin
				if Id /= Lower_Case_Id then
					declare
						Lower_Case_Name : constant String := Management.Name (Object, Lower_Case_Id);
					begin
						Ada.Directories.Symbolic_Link (Name, Lower_Case_Name, Overwrite => False);
					exception
						when Ada.Directories.Use_Error =>
							if not Ada.Directories.Exists (Lower_Case_Name) then
								raise;
							end if;
					end;
				end if;
			end;
		end;
	end Add;
	
	procedure Query (
		Object : in out Manager;
		Id : in String;
		Password : in Password_Digest;
		Remote_Addr : in String;
		Remote_Host : in String;
		Time : in Ada.Calendar.Time;
		Info : out User_Info)
	is
		Name : constant String := Management.Name (Object, Id);
	begin
		Load (Name, Info);
		if Info.Password /= Password then
			raise Password_Mismatch;
		end if;
		Info.Last_Remote_Addr := +Remote_Addr;
		Info.Last_Remote_Host := +Remote_Host;
		Info.Last_Time := Time;
		Save (Name, Info);
	end Query;
	
	procedure Query (
		Object : in out Manager;
		Id : in String;
		Watching : out Watching_Syndications.List)
	is
		Info : User_Info;
	begin
		Load (Name (Object, Id), Info);
		Watching := Info.Watching;
	end Query;
	
	procedure Update (
		Object : in out Manager;
		Id : in String;
		Info : in out User_Info) is
	begin
		Save (Name (Object, Id), Info);
	end Update;
	
end Birdeyes.Users.Management;
